/***********************************************************************
 * File: conf_file_loader.c
 * 	(Extracted from
 *	   "$Id: conf_file_loader.c,v 1.2 2009/05/21 12:57:26 maoyam Exp maoyam $")
 *
 * Abstracts:
 *  This C source implements "Configuration file loader" API.
 *
 * Copyright (C) 2005-2007 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

#include "conf_file_loader.h" /* definitions for API */


/* define the extension of "INI" files */
#define INI_FILE_EXT	".INI"
#define INI_FILE_EXT_2	".ini"


/* get the next describe line (skip the comment lines and blank lines)
 */
static char *get_next_row(char *cp, char *cmntp)
{
	char	*np, *result = NULL, *commentp = cmntp;

	/* until reached the end-of-file... */
	while (*cp != '\0') {
		char	*dstp = NULL, *skipedp = NULL;
		int		next_line = 0; /* for controlling loop */

		cmntp	= commentp;	/* restore the comment string */

		/* skip the spaces in the head of a line... */
		while ((*cp != '\0') && isspace(*cp)) {	++cp;	}

		/* skip the blank lines */
		if (strncmp(cp, "\r\n", 2) == 0) {
			cp	+= 2;
			continue;
		}
		if ((*cp == '\r') || *cp == '\n') {
			++cp;
			continue;
		}

		/* skip the comment lines... */
		while (*cmntp != '\0') {
			/* test: the head character is matched to a comment character?
			 */
			if (*cp == *cmntp) {
				/* test: the C style comment isn't closed? */
				if (((*cmntp == '/')
				&& strncmp(cp, "/*", 2) == 0)
				&& (cp = strstr(cp + 2, "*/")) == NULL) {
					return	NULL; /* all lines are comment */
				}

				/* test: the comment line ends with '\r" or "\r\n"? */
				if ((np = strchr(cp, '\r')) != NULL) {
					if (*(np + 1) == '\n') {
						cp	= np + 2;
					}
					else {
						cp	= np + 1;
					}
				}
				else if ((np = strchr(cp, '\n'))
							!= NULL) {
					/* the comment line ends with '\n' */
					cp	= np + 1;
				}
				else {	/* this comment line is last. */
					return	NULL;
				}

				/* set the loop-control variable to continue the outer loop
				 */
				next_line = 1; break;
			}

			++cmntp; /* try to next comment character */
		}

		/* test: the loop-control variable is set to continue
		 * the outer loop?
		 */
		if (next_line != 0) {
			continue; /* skip the comment line, and try to the next line */
		}

		/* test: the address of the first character of the line
		 * isn't set, yet?
		 */
		if (result == NULL) {
			result	= cp; /* set the address to result */
		}

		/* search the end-of-line */
		if (((np = strchr(cp, '\r')) != NULL) && *(np - 1) == '\\') {
			/* the end-of-line is "\\\r". concatenate it and the next line
			 */
			dstp	= np - 1;
			skipedp	= (np + ((*(np + 1) == '\n')? 2: 1));
			strcpy(dstp, skipedp);
			cp		= np - 1;
		}
		else if (((np = strchr(cp, '\n')) != NULL) && *(np - 1) == '\\') {
			/* the end-of-line is "\\\n". concatenate it and the next line
			 */
			dstp	= np - 1;
			skipedp	= np + 1;
			strcpy(dstp, skipedp);
			cp		= np - 1;
		}
		else {	/* this line is last one. */
			break;
		}
	}

	return	result; /* the address of the first character of this line */
}


/* trim the spaces in the left-hand-side of the string */
static char *trim_left_spaces(char *cp)
{
	/* skip the spaces in the head of the string... */
	while ((*cp != '\0') && isspace(*cp)) {	++cp;	}

	return	cp; /* the address of the first non-space character */
}


/* trim the spaces in the right-hand-side of the string */
static char *trim_right_spaces(char *kp)
{
	char	*sp = kp; /* save the address of the string */
	size_t	len = strlen(kp); /* get the length of the string */

	/* test: len is greater than zero? */
	if (len != 0) {
		/* eliminate the spaces in the right-hand-side of the string */
		for (kp += (len - 1); kp > sp; --kp) {
			/* test: the character is space? */
			if (isspace(*kp)) {
				*kp	= '\0'; /* eliminate it */
			}
			else {	/* find the non-space character */
				break; /* end of processing */
			}
		}
	}

	return	sp; /* the address of the string */
}


/* decode the escape sequence of the string for Java-property style */
static char *deescape(char *cp)
{
	char	*savep	= cp; /* save the address of the string */

	/* until reached the end of string... */
	while (*cp != '\0') {
		/* test: the character is escape character '\\'? */
		if (*cp == '\\') {
			/* test: the last character is '\\'? */
			if (*(cp + 1) == '\0') {
				/* it isn't escape sequence. thus, eliminate it */
				*cp = '\0'; break; /* end of processing */
			}

			/* which escape sequence? */
			switch (*(cp + 1)) {
			case 'r': /* '\r' means LF (line feed) */
				*cp = '\r'; break;

			case 'n': /* "\n" means CR (carriage return) */
				*cp = '\n'; break;

			case 'a': /* '\a' means BELL */
				*cp = '\a'; break;

			case 't': /* '\t' means TAB */
				*cp = '\t'; break;

			case 'b': /* '\b' means BACK-SPACE */
				*cp = '\b'; break;

			default: /* otherwise, escape the next code */
				*cp = *(cp + 1); break;
			}

			/* move the next of the escape sequence */
			cp	= strcpy(cp + 1, cp + 2);
		}
		else {	/* ordinal characters */
			++cp; /* ignore it */
		}
	}

	return	savep; /* the address of the decoded string */
}


/* compare keys of two row structures in the Configuration_t */
static int cmp_row_keys(const void *row1, const void *row2)
{
	return	strcmp(((Conf_row_t *)row1)->key, ((Conf_row_t *)row2)->key);
}


/*
 * This API loads the configuration file, that's style is DOS's
 * INI-files or Java's property-files, and the builds the
 * Configuration_t structure.
 *
 * arguments:
 *	path is pointing to the path of the configuration file.
 *	comment_chars is pointing to the comment characters.
 *	val_delimit indicates the delimit character of the values.
 *
 * returns:
 *	non-NULL: the Configuration_t structure that is loaded the file.
 *	NULL: several failure is happened. errno indicates the details.
 *		errno : means...
 *		EINVAL: illegal arguments
 *		EACCES: the file isn't readable access mode
 *		EIO:	failed to read the file to allocated area
 *		ENOMEM:	insufficient memory
 */
Configuration_t *load_conf_file(char *path
								, char *comment_chars
								, char val_delimit
								)
{
	Configuration_t	*conf;
	struct stat		st;
	size_t			conf_sz;
	char			*cp, *contents = NULL;
	FILE			*fp = NULL;
	int				is_INI_file = 0;

	errno	= 0; /* initialize errno */

	/* initialize the stat structure */
	memset(&st, (int)'\0', sizeof(st));

	/* test: illegal arguments, or access error? */
	if ((path == NULL) || comment_chars == NULL) {
		errno = EINVAL; return NULL; /* FAILED: illegal arguments */
	}
	else if (stat(path, &st) != 0) {
		return NULL; /* FAILED: fail access */
	}
	else if (!((S_ISREG(st.st_mode)) || S_ISLNK(st.st_mode))) {
		/* FAILED: file isn't regular or symbolic link */
		errno = EINVAL; return NULL;
	}
	else if ((st.st_mode & (S_IRUSR | S_IRGRP | S_IROTH)) == 0) {
		errno = EACCES; return NULL; /* FAILED: can't read the file */
	}
	else if ((contents = (char *)malloc((size_t)(st.st_size + 1))) == NULL)
	{	return NULL; /* FAILED: insufficient memory */	}
	else if ((fp = fopen(path, "r")) == NULL) {
		free(contents); contents = NULL;
		return NULL; /* FAILED: can't open the file */
	}
	else if (fread(contents, sizeof(*contents), (size_t)st.st_size, fp)
													!= (size_t)st.st_size) {
		(void)fclose(fp); fp	= NULL;
		free(contents); contents = NULL;
		errno = EIO; return NULL; /* FAILED: can't load the file */
	}
	else {	/* read in the file to the 'contents' allocated area */
		(void)fclose(fp); fp	= NULL; /* close the file */
		contents[st.st_size]	= '\0'; /* set the EOF */

		/* determine the file is DOS's INI-file style, or not */
		is_INI_file	= (((strstr(path + (strlen(path) - strlen(INI_FILE_EXT))
								, INI_FILE_EXT
								) != NULL)
						|| strstr((path +
									(strlen(path) - strlen(INI_FILE_EXT_2))
									)
								, INI_FILE_EXT_2
								) != NULL)?		1: 0
						);
	}

	/* allocate the Configuration_t structure and initialize it */
	conf_sz	= sizeof(*conf) - sizeof(conf->rows[0]);
	if ((conf = (Configuration_t *)malloc(conf_sz)) == NULL) {
		free(contents); contents = NULL;
		errno = ENOMEM; return NULL; /* FAILED: insufficient memory */
	}
	memset(conf, (int)'\0', conf_sz);
	conf->top_addr	= contents;
	conf->file_alloc_sz	= (size_t)(st.st_size + 1);
	conf->rows_alloc_sz	= 0;

	/* parse the contents of the file... */
	for (cp = contents; cp < contents + st.st_size; ++cp) {
		char		*lp;
		Conf_row_t	*rows;

		/* get the non-comment line */
		if (((cp = get_next_row(cp,comment_chars)) == NULL)
		|| cp >= contents + st.st_size) {
			/* the last line is comment line. */
			break; /* end of the processing */
		}

		/* search the end of the line */
		if ((lp = strchr(cp, '\r')) != NULL) {
			*lp	= '\0';
		}
		else if ((lp = strchr(cp, '\n')) != NULL) {
			*lp	= '\0';
		}
		else {	/* this line is last one */
			lp	= cp + strlen(cp);
		}

		/* extend the Configuration_t structure */
		conf_sz	+= sizeof(conf->rows[0]);
		if ((conf = (Configuration_t *)realloc(conf, conf_sz))
							== NULL) {
			free(memset(contents, (int)'\0', (size_t)(st.st_size + 1)));
			contents	= NULL;
			errno = ENOMEM; return NULL; /* FAILED: * insufficient memory */
		}
		conf->rows_alloc_sz	+= sizeof(conf->rows[0]);

		/* cut the key string in this line */
		rows		= conf->rows + conf->rows_n;
		++(conf->rows_n);
		rows->key	= cp;
		for (++cp; ((cp < lp) && *cp != '='); ++cp) { ; }
		*cp			= '\0';
		(void)trim_right_spaces(rows->key);
		if (is_INI_file == 0) {
			/* non-INI file can contain escape sequences */
			(void)deescape(rows->key);
		}
			
		/* skip the spaces in the head of the values string */
		rows->vals_n	= 0;
		rows->vals		= NULL;
		for (++cp; ((cp < lp) && ((*cp == '=') || isspace(*cp))); ++cp)
		{	;	}

		/* test: the values string isn't empty? */
		if (cp < lp) {
			char	*vp;

			/* until the last of the values string... */
			while (cp < lp) {
				/* test: enclosed with double-quotes? */
				if ((*cp == '"') && (vp = strchr(cp + 1, '"')) != NULL) {
					/* eliminate the quotation */
					++cp; *vp = '\0';
				}
				else {	/* not quoted */
					vp	= cp;
				}

				/* test: this value has next value? */
				if ((vp = strchr(vp + 1, val_delimit)) == NULL) {
					vp	= lp;
				}
				else {	/* this's the last value */
					*vp	= '\0';
				}

				/* extend the value pointer vector */
				if ((rows->vals = (char **)realloc(rows->vals
												, (rows->vals_n + 1) *
													sizeof(rows->vals[0])
												)) == NULL) {
					/* FAILED: extend the vector */
					size_t	i;

					/* release the value pointer vector */
					for (i = 0; i < conf->rows_n; ++i) {
						if (conf->rows[i].vals != NULL) {
							free(conf->rows[i].vals);
							conf->rows[i].vals	= NULL;
						}
					}

					/* release the file read in area
					 * and the Configuration_t structure
					 */
					free(memset(conf->top_addr, (int)'\0'
								, (size_t)(st.st_size + 1)
						))
					; conf->top_addr	= NULL
					; contents			= NULL
					; free(conf); conf = NULL
					;

					/* FAILED: insufficient memory */
					errno = ENOMEM; return NULL;
				}

				/* set this value to the vector.
				 * (if the file isn't INI-file style, decode the escape
				 * sequences and trim the left-hand-side spaces)
				 */
				rows->vals[rows->vals_n]	=
						(is_INI_file == 0)?	trim_left_spaces(deescape(cp))
						:					cp
						;

				/* increment the number of values, values vector allocated
				 * size and contents pointer (cp)
				 */
				++(rows->vals_n);
				rows->vals_alloc_sz =
						sizeof(rows->vals[0]) * rows->vals_n
						;
				cp	= (vp == lp)? vp: vp + 1;
			}
		}
	}

	/* sort the rows of Configuration_t, using the keys */
	qsort(conf->rows, conf->rows_n, sizeof(conf->rows[0]), cmp_row_keys);

	return	conf; /* the Configuration_t that loads the file */
}


/*
 * This API get the address of Conf_row_t slot in the Configure_t,
 * that's key is matched the specified key string.
 *
 * arguments:
 *	conf is pointing the returns of the load_conf_file() API.
 *	key is pointing to the key string.
 *
 * returns:
 *	non-NULL: pointing the Conf_row_t slot that is matched the key
 *		string.
 *	NULL: the Conf_row_t is empty in the conf, or illegal arguments
 *		errno : means...
 *		zero  : the Conf_row_t slot of conf is empty.
 *		EINAVL: illegal arguments
 */
Conf_row_t *get_row_matched_key(Configuration_t *conf, char *key)
{
	Conf_row_t	key_, *result = NULL;

	errno	= 0; /* initialize errno */

	/* test: illegal arguments? */
	if (conf == NULL || key == NULL) {
		errno = EINVAL; return NULL; /* FAILED: illegal arguments */
	}

	/* test: the Conf_row_t slot in conf isn't empty? */
	if (conf->rows_n > 0) {
		/* set the key specifying Conf_row_t datum */
		memset(&key_, (int)'\0', sizeof(key_));
		key_.key	= key;

		/* search the slot in the conf, using the key string */
		result	= (Conf_row_t *)bsearch(&key_
										, conf->rows
										, conf->rows_n
										, sizeof(key_)
										, cmp_row_keys
										);

		/* rewind the same key rows... */
		while ((result > conf->rows)
		&& strcmp(result->key, (result - 1)->key) == 0
		) {	--result;	}
	}
	else {	/* the Conf_row_t in conf is empty */
		result	= NULL;
	}

	return	result;	/* the Conf_row_t slot address, or NULL */
}


/*
 * This API find the value slot in the value vector, that is pointed
 * from the Conf_row_t in conf that is matched with key string. And
 * outputs the index of the values vector.
 *
 * arguments:
 *	conf is the returns of load_conf_file() API.
 *	key is pointing to the key string.
 *	value is pointing the finding value string.
 *	val_idx outputs the index of the values vector.
 *
 * returns:
 *	non-NULL: address of the found value string in the vector
 *	NULL: not-found the value in the vector, or illegal arguments
 *		errno : means...
 *		zero  : not found the value in the values vector
 *		EINVAL: illegal arguments
 */
char *find_value_matched_key(Configuration_t	*conf
							, char				*key
							, char				*value
							, size_t			*val_idx
							)
{
	Conf_row_t	*row;
	char		*result;

	errno	= 0; /* initialize errno */

	/* test: illegal arguments? */
	if ((value == NULL) || val_idx == NULL) {
		errno = EINVAL; return NULL; /* FAILED: illegal arguments */
	}

	/* get the Conf_row_t slot that is matched the key string */
	if ((row = get_row_matched_key(conf, key)) != NULL) {
		/* search the index in the vector, that equals the value string */
		for (;;) {
			for (*val_idx = 0, result = NULL
			; *val_idx < row->vals_n
			; ++(*val_idx)
			) {
				if (strcmp(value, row->vals[*val_idx]) == 0) {
					/* found the value index */
					result = row->vals[*val_idx]; break;
				}
			}
			if (((result != NULL)
			|| row + 1 >= conf->rows + conf->rows_n)
			|| strcmp((row + 1)->key, key) != 0
			) {	break;	}
			++row;
		}
	}
	else {	/* not found the Conf_row_t slot matched key string */
		result	= NULL;	/* not found */
	}

	/* the address of the value in vector, or NULL (not found) */
	return	result;
}


/*
 * This API release the allocated memories for Configuration_t structure
 *
 * arguments:
 *	conf is the returns of load_conf_file() API.
 *
 * returns:
 *	nothing
 */
void destroy_conf(Configuration_t *conf)
{
	size_t	i;

	/* test: the conf isn't NULL? */
	if (conf != NULL) {
		/* release the values vector allocated areas... */
		for (i = 0; i < conf->rows_n; ++i) {
			/* clear the memory, before release it */
			memset(conf->rows[i].vals
					, (int)'\0'
					, (conf->rows[i].vals_n * sizeof(conf->rows[0].vals[0]))
					);
			free(conf->rows[i].vals);
		}

		/* release the file contents area */
		memset(conf->top_addr, (int)'\0', conf->file_alloc_sz);
		free(conf->top_addr);

		/* release the Configuration_t structure */
		memset(conf, (int)'\0'
				, (sizeof(*conf)
						+ ((conf->rows_n - 1) * sizeof(conf->rows[0])))
				);
		free(conf);
	}

	return; /* nothing */
}
